/*
 * dfu-programmer
 *
 * $Id: dfu.c,v 1.3 2006/06/20 06:28:04 schmidtw Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <usb.h>
#include "dfu.h"

/* DFU commands */
#define DFU_DETACH      0
#define DFU_DNLOAD      1
#define DFU_UPLOAD      2
#define DFU_GETSTATUS   3
#define DFU_CLRSTATUS   4
#define DFU_GETSTATE    5
#define DFU_ABORT       6

#define INVALID_DFU_TIMEOUT -1

static int dfu_timeout = INVALID_DFU_TIMEOUT;
static unsigned short transaction = 0;
static unsigned short sent = 0;

static int dfu_debug_level = 1;

void dfu_init( const int timeout )
{
    if( timeout > 0 ) {
        dfu_timeout = timeout;
    } else {
        if( 0 != dfu_debug_level )
            fprintf( stderr, "dfu_init: Invalid timeout value.\n" );
    }
}

static int dfu_verify_init( const char *function )
{
    if( INVALID_DFU_TIMEOUT == dfu_timeout ) {
        if( 0 != dfu_debug_level )
            fprintf( stderr,
                     "%s: dfu system not property initialized.\n",
                     function );
        return -1;
    }

    return 0;
}

void dfu_debug( const int level )
{
    dfu_debug_level = level;
}

void printdump(char * buffer, int len) 
{
char n;
		fprintf(stderr, "\n%03x:", 0);
    for (n=0;n<len;n++) {
      fprintf( stderr, "%02x",(int) (buffer[n] & 0xFF));
      if ((n+1) % 8  ==0) fprintf(stderr, " ");
      if ((n+1) % 32 ==0 && (n+1<len)) fprintf(stderr, "\n%03x:", n+1);
    }
    fprintf(stderr, "\n");
}


/*  NOT USED !
 *  DFU_DETACH Request (DFU Spec 1.0, Section 5.1)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *  timeout   - the timeout in ms the USB device should wait for a pending
 *              USB reset before giving up and terminating the operation
 *
 *  returns 0 or < 0 on error
 */
int dfu_detach( struct usb_dev_handle *device,
                const unsigned short interface,
                const unsigned short timeout )
{
    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    return usb_control_msg( device,
        /* bmRequestType */ USB_ENDPOINT_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
        /* bRequest      */ DFU_DETACH,
        /* wValue        */ timeout,
        /* wIndex        */ interface,
        /* Data          */ NULL,
        /* wLength       */ 0,
                            dfu_timeout );
}

// Get a buffer
int dfu_ctrl_read(struct usb_dev_handle *device, unsigned char bmRequestType, unsigned char bRequest, unsigned short wValue, unsigned short wIndex, char *buffer, unsigned short wLength)
{
    int result;

		memset((void*) buffer, 0, wLength);

	  result = usb_control_msg( device,
                              bmRequestType, 
                              bRequest,
                              wValue,
                              wIndex,
                              buffer,
                              wLength,
                              dfu_timeout );
    //Ok ?
    if (result==wLength)
      result=0;
      
    fprintf( stderr, "%s: %02x%02x >%x err=%d %s\n", __FUNCTION__, bmRequestType, bRequest, result, 0-result, (result==0)? "" : usb_strerror());
		printdump(buffer,wLength);
		
		usleep(50000);
		return result;
}

// Get a buffer
int dfu_ctrl_get(struct usb_dev_handle *device, unsigned char bRequest, unsigned short wValue, unsigned short wIndex, char *buffer, unsigned short wLength)
{
    int result;

		memset((void*) buffer, 0, wLength);

	  result = usb_control_msg( device,
                              0xC0, //bmRequestType, 
                              bRequest,
                              wValue,
                              wIndex,
                              buffer,
                              wLength,
                              dfu_timeout );
    //Ok ?
    if (result==wLength)
      result=0;
      
    fprintf( stderr, "%s: C0%02x >%x err=%d %s\n", __FUNCTION__, bRequest, result, 0-result, (result==0)? "" : usb_strerror());
		printdump(buffer,wLength);
		
		usleep(50000);
		return result;
}

// Write a buffer
int dfu_ctrl_set(struct usb_dev_handle *device, unsigned char bRequest, unsigned short wValue, unsigned short wIndex, char *buffer, unsigned short wLength)
{
    int result;

	  result = usb_control_msg( device,
                              0x40, //bmRequestType, 
                              bRequest,
                              wValue,
                              wIndex,
                              buffer,
                              wLength,
                              dfu_timeout );
    //Ok ?
    if (result==wLength)
      result=0;
      
    fprintf( stderr, "%s: 40%02x >%x err=%d %s\n", __FUNCTION__, bRequest, result, 0-result, (result==0)? "" : usb_strerror());
		printdump(buffer,wLength);
		
		usleep(5000);
		return result;
}

// Write a buffer
int dfu_ctrl_class(struct usb_dev_handle *device, unsigned short wValue, unsigned short wIndex, char *buffer, unsigned short wLength)
{
    int result;
    
	  result = usb_control_msg( device,
                              0x21, //bmRequestType, 
                              0x09,
                              wValue,
                              wIndex,
                              buffer,
                              wLength,
                              dfu_timeout );

    if (result==wLength)
      result=0;
      
    fprintf( stderr, "%s: 2109 in  %x >%x err=%d %s\n", __FUNCTION__, buffer[0], result, 0-result, (result==0)? "" : usb_strerror());		
		
		usleep(5000);
		return result;
}

// Exec command
int dfu_ctrl(struct usb_dev_handle *device, unsigned char bRequest, unsigned short wValue, unsigned short wIndex)
{
    int result;

	  result = usb_control_msg( device,
                              0x40, // bmRequestType, 
                              bRequest,
                              wValue,
                              wIndex,
                              0,
                              0,
                              dfu_timeout );
    
   	fprintf( stderr, "%s: 40%02x >%x err=%d %s\n", __FUNCTION__, bRequest, result, 0-result, (result==0)? "" : usb_strerror());
    
		usleep(5000);
		return result;
}

int dfu_before_download( struct usb_dev_handle *device,
                        const unsigned short interface,
                        struct dfu_status *status )
{
    char buffer[2048];
    int result;

		//Class Interface
	  result = usb_control_msg( device,
                              0x21, // bmRequestType, 
                              0x0A,
                              0,
                              0,
                              0,
                              0,
                              dfu_timeout );
    fprintf( stderr, "%s: 210A 0 0 0 0 >%x err=%d %s\n", __FUNCTION__, result, 0-result, (result==0)? "" : usb_strerror());
    
		memset(buffer,0,0x200);
	  result = dfu_ctrl_get( device,
          /* bRequest      */ 0x20, 
          /* wValue        */ 0,
          /* wIndex        */ 0x0,
          /* Data          */ buffer,
          /* wLength       */ 0x200
                          );

    fprintf( stderr, "%s: F C020 0 0 B 200 >%x err=%d %s\n", __FUNCTION__, result, 0-result, (result==0)? "" : usb_strerror());
		printdump(&buffer[0],0x200);
				
status->bStatus = DFU_STATUS_OK;
status->bState  = STATE_DFU_IDLE;

		return result;
}

/*
 *  DFU_DNLOAD Request (DFU Spec 1.0, Section 6.1.1)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *  length    - the total number of bytes to transfer to the USB
 *              device - must be less than wTransferSize
 *  data      - the data to transfer
 *
 *  returns the number of bytes written or < 0 on error
 */
int dfu_download( struct usb_dev_handle *device,
                  const unsigned short interface,
                  const unsigned short length,
                  char* data )
{
    int status;
    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    /* Sanity checks */
    if( (0 != length) && (NULL == data) ) {
        if( 0 != dfu_debug_level )
            fprintf( stderr,
                     "%s: data was NULL, but length != 0\n",
                     __FUNCTION__ );
        return -1;
    }

    if( (0 == length) && (NULL != data) ) {
        if( 0 != dfu_debug_level )
            fprintf( stderr,
                     "%s: data was not NULL, but length == 0\n",
                     __FUNCTION__ );
        return -2;
    }

    status = usb_control_msg( device,
          /* bmRequestType */ 0x40,//USB_ENDPOINT_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ 0,//transfert,
          /* wValue        */ 0, //transaction++, //inutile
          /* wIndex        */ sent, //offset
          /* Data          */ data,
          /* wLength       */ length,
                              dfu_timeout );
    if( status < 0 ) {
        fprintf( stderr, "%s: error %d\n", __FUNCTION__, status );
    } else {
    	sent+=length;
    }

    return status;
}

int fill_bloc(char * buffer)
{
	int n=-1;
	n++; buffer[n]=0x09;
	n++; buffer[n]=0xB1;
	n++; buffer[n]=0xFE;
	return n;
}

int fill_first(char * buffer)
{
	int n=-1;
	n++; buffer[n]=0x0B; 	n++; buffer[n]=0x16;
	n++; buffer[n]=0x0C; 	n++; buffer[n]=0xD9;
	n++; buffer[n]=0x0D; 	n++; buffer[n]=0x99;
	n++; buffer[n]=0x16;
	n++; buffer[n]=0x5F;
	n++; buffer[n]=0x20;
	n++; buffer[n]=0x40;
	n++; buffer[n]=0x2B; 	n++; buffer[n]=0x33;
	n++; buffer[n]=0x2C; 	n++; buffer[n]=0x20;
	n++; buffer[n]=0x35;
	n++; buffer[n]=0x83;
	n++; buffer[n]=0x38;
	n++; buffer[n]=0x88;
	n++; buffer[n]=0x3B;
	n++; buffer[n]=0x00;
	n++; buffer[n]=0x49;
	n++; buffer[n]=0xA0;
	n++; buffer[n]=0x4C;
	n++; buffer[n]=0xFB;
	n++; buffer[n]=0x4D;
	n++; buffer[n]=0x93;
	n++; buffer[n]=0x5B;
	n++; buffer[n]=0x70;
	n++; buffer[n]=0x86;
	n++; buffer[n]=0x04;
	n++; buffer[n]=0x87; 
	n++; buffer[n]=0x01;
	n++; buffer[n]=0x89;
	n++; buffer[n]=0xC8;
	n++; buffer[n]=0x93;
	n++; buffer[n]=0xDA;
	n++; buffer[n]=0x9C;
	n++; buffer[n]=0x57;
	n++; buffer[n]=0xA6; 	n++; buffer[n]=0x9D;
	n++; buffer[n]=0xA7; 	n++; buffer[n]=0x01;
	n++; buffer[n]=0xA8; 	n++; buffer[n]=0xD2;
	n++; buffer[n]=0x19;
	n++; buffer[n]=0x36;
	n++; buffer[n]=0xFE;
	
	return n;
}

int fill_second(char * buffer)
{
	int n=-1;
	n++; buffer[n]=0x0E;  n++; buffer[n]=0x32; 
	n++; buffer[n]=0x0F;  n++; buffer[n]=0xA2; 
	n++; buffer[n]=0x10;  n++; buffer[n]=0x8F; 
	n++; buffer[n]=0x11;  n++; buffer[n]=0x60; 
	n++; buffer[n]=0x16;
	n++; buffer[n]=0x5F;
	n++; buffer[n]=0x2B;
	n++; buffer[n]=0x43;
	n++; buffer[n]=0x44;  n++; buffer[n]=0xC0; 
	n++; buffer[n]=0x45;  n++; buffer[n]=0x01; 
	n++; buffer[n]=0x46;  n++; buffer[n]=0x9A; 
	n++; buffer[n]=0x49;
	n++; buffer[n]=0xB0;
	n++; buffer[n]=0x5C;  n++; buffer[n]=0x00; 
	n++; buffer[n]=0x5D;  n++; buffer[n]=0x02; 
	n++; buffer[n]=0x6A;  n++; buffer[n]=0x47; 
	n++; buffer[n]=0x6B;  n++; buffer[n]=0x84; 
	n++; buffer[n]=0x6C;  n++; buffer[n]=0x99; 
	n++; buffer[n]=0x6D;  n++; buffer[n]=0x4E; 
	n++; buffer[n]=0x6E;  n++; buffer[n]=0x55; 
	n++; buffer[n]=0x6F;  n++; buffer[n]=0xE8; 
	n++; buffer[n]=0x70;  n++; buffer[n]=0x82; 
	n++; buffer[n]=0x88;
	n++; buffer[n]=0x7E;
	n++; buffer[n]=0x8A;
	n++; buffer[n]=0x38;
	n++; buffer[n]=0x95;
	n++; buffer[n]=0x03;
	n++; buffer[n]=0x4D;
	n++; buffer[n]=0x93;
	n++; buffer[n]=0xA6;  n++; buffer[n]=0x9D; 
	n++; buffer[n]=0xA7;  n++; buffer[n]=0x01; 
	n++; buffer[n]=0xA8;  n++; buffer[n]=0xD2; 
	n++; buffer[n]=0x09;  n++; buffer[n]=0xF3;
	n++; buffer[n]=0xFE;

	return n;
}

int dfu_after_download( struct usb_dev_handle *device,
                        const unsigned short interface,
                        struct dfu_status *status )
{
    char buffer[2048];
    int result;

		usleep(100*1000);
		usleep(100*1000);
		usleep(100*1000);
		usleep(100*1000);
    
    //#187 reset ?
		result = dfu_ctrl(device, 1,	0,	1);

		usleep(100*1000);
		usleep(100*1000);
		usleep(100*1000);
		usleep(100*1000);
		
    fprintf( stderr, "%s\n", __FUNCTION__);
		
		//                    		Req		wValue	wIndex
		//#190 EEPROM ?
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0206);
		if (result==-19) {
		  
		  result = dfu_ctrl(device, 1,	0,	1);
		  
		  result = usb_reset(device);
			if (result < 0 && result != -19) { //-ENODEV) {
				fprintf(stderr, "error resetting after download: %s\n", usb_strerror());
			}

			result = usb_set_altinterface(device, 0);
	        fprintf( stderr, "%s: altinterface 0 res=%x %s\n", __FUNCTION__, 0-result, (result==0)? "" : usb_strerror());		
	    result = usb_release_interface(device, 1);
	        fprintf( stderr, "%s: release 1 res=%x %s\n", __FUNCTION__, 0-result, (result==0)? "" : usb_strerror());
  	  result = usb_release_interface(device, 0);
  	      fprintf( stderr, "%s: release 0 res=%x %s\n", __FUNCTION__, 0-result, (result==0)? "" : usb_strerror());
		  result = usb_claim_interface(device, 1);
		      fprintf( stderr, "%s: claim 1 res=%x %s\n", __FUNCTION__, 0-result, (result==0)? "" : usb_strerror());
//	    result = usb_claim_interface(device, 0);
		}//
		if (result==-19) return result;

		//Switch On LED (and maybe other things)			
		result = usb_control_msg( device,
          /* bmRequestType */ 0x40, //USB_ENDPOINT_IN, //0x21,//USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ 0x04, //DFU_GETSTATUS,
          /* wValue        */ 1,
          /* wIndex        */ 8,
          /* Data          */ 0,
          /* wLength       */ 0,
                              dfu_timeout );
                              
    fprintf( stderr, "%s: F 4004 1 8 0 0 res=%x %s\n", __FUNCTION__, 0-result, (result==0)? "" : usb_strerror());
		usleep(5000);

#if 1	
		//#193 EEPROM DUMP ? 0xA0 is EEPROM 24C02C I2C addr
		result = dfu_ctrl_get(device, 0x20,	0x0000,	0x00a0,	buffer, 0x1a); //return FFFF FFFF...
		result = dfu_ctrl_get(device, 0x20,	0x0000,	0x01a0,	buffer, 0x1a); //return 1689 ..
		
		result = dfu_ctrl(device, 0x04,	0x0001,	0x0008);
		
		result = dfu_ctrl(device, 0x30,	0x0000, 0xfff0);
		result = dfu_ctrl(device, 0x30,	0x0006, 0xfff1);
		result = dfu_ctrl(device, 0x30,	0x0001, 0xfff2);
		result = dfu_ctrl(device, 0x30,	0x866b, 0xffff);
		
		//IRKeys table ??
		result = dfu_ctrl(device, 0x30, 0x0710, 0x13); //220 > 1310 07
		result = dfu_ctrl(device, 0x30, 0x070A, 0x11); //223 > 110A 07
		result = dfu_ctrl(device, 0x30, 0x0711, 0x1C); //226 > 1C11 07
		result = dfu_ctrl(device, 0x30, 0x0713, 0x12); //229
		result = dfu_ctrl(device, 0x30, 0x071E, 0x01); //232
		result = dfu_ctrl(device, 0x30, 0x071F, 0x02); //235
		result = dfu_ctrl(device, 0x30, 0x0720, 0x03); //238
		result = dfu_ctrl(device, 0x30, 0x0714, 0x0D); //241
		result = dfu_ctrl(device, 0x30, 0x0721, 0x04); //244
		result = dfu_ctrl(device, 0x30, 0x0722, 0x05); //247
		result = dfu_ctrl(device, 0x30, 0x0723, 0x06); //250
		result = dfu_ctrl(device, 0x30, 0x070B, 0x19); //253
		result = dfu_ctrl(device, 0x30, 0x0724, 0x07); //256
		result = dfu_ctrl(device, 0x30, 0x0725, 0x08); //259
		result = dfu_ctrl(device, 0x30, 0x0726, 0x09); //262
		result = dfu_ctrl(device, 0x30, 0x071B, 0x10); //265
		result = dfu_ctrl(device, 0x30, 0x070D, 0x0A); //268
		result = dfu_ctrl(device, 0x30, 0x0727, 0x00); //271
		result = dfu_ctrl(device, 0x30, 0x0718, 0x0B); //274
		result = dfu_ctrl(device, 0x30, 0x0719, 0x0C); //277
		result = dfu_ctrl(device, 0x30, 0x0708, 0x1B); //280
		result = dfu_ctrl(device, 0x30, 0x0715, 0x1D); //283
		result = dfu_ctrl(device, 0x30, 0x0707, 0x15); //286
		result = dfu_ctrl(device, 0x30, 0x070F, 0x18); //289
		result = dfu_ctrl(device, 0x30, 0x0712, 0x0E); //292
		result = dfu_ctrl(device, 0x30, 0x0716, 0x1E); //295
		result = dfu_ctrl(device, 0x30, 0x0705, 0x0F); //298
		result = dfu_ctrl(device, 0x30, 0x0709, 0x1A); //301
		
		buffer[0]=1;
    dfu_ctrl_class(device, 0x0200, 0x00, buffer, 1);
    
#endif		
		//try eeprom dump
		memset((void*) buffer, 0, 0x200);
		memset((void*) buffer, 0x00, 0x200);
#if 0		
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0008);
		usleep(1000000);
		result = dfu_ctrl_get(device, 0x01,	0x0008,	0x0002, buffer, 0x01); //return 1689 ..		
		
		result = dfu_ctrl(device, 0x04,	0x0001,	0x0008);
				usleep(1000000);
		result = dfu_ctrl_get(device, 0x01,	0x0008,	0x0002, buffer, 0x01); //return 1689 ..		

		result = dfu_ctrl(device, 0x04,	0x0000,	0x0008);
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0006);
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0200);
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0208);
		result = dfu_ctrl(device, 0x04,	0x0000,	0x0206);
#endif
		/*
		dfu_ctrl_get(device, 0x03, 0x0000, 0xFF01, buffer, 0x01); //596
    dfu_ctrl    (device, 0x04, 0x0001, 0x0009);               //599
    dfu_ctrl    (device, 0x03, 0x0001, 0xFF34);               //602
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF09, buffer, 0x01); //605
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF0A, buffer, 0x01); //608
    */
    dfu_ctrl    (device, 0x04, 0x0000, 0x0206);               //611
    dfu_ctrl_get(device, 0x20, 0x0000, 0x00A0, buffer, 0x1A); //614 > FFFF...
    dfu_ctrl_get(device, 0x20, 0x0000, 0x01A0, buffer, 0x1A); //617 > 1689...
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFA06, buffer, 0x01); //620 > 0x18
    dfu_ctrl    (device, 0x04, 0x0001, 0x000B);               //623
    dfu_ctrl    (device, 0x04, 0x0000, 0x000A);               //626
    dfu_ctrl    (device, 0x03, 0x0003, 0xFF04);               //629
    dfu_ctrl    (device, 0x03, 0x0058, 0xFF67);               //632
    dfu_ctrl    (device, 0x03, 0x0030, 0xFF19);               //635
    dfu_ctrl    (device, 0x03, 0x0018, 0xFF05);               //638
    dfu_ctrl    (device, 0x04, 0x0001, 0x0008);               //641
    dfu_ctrl    (device, 0x03, 0x0003, 0xFF04);               //644
    dfu_ctrl    (device, 0x03, 0x0058, 0xFF67);               //647
    dfu_ctrl    (device, 0x03, 0x0030, 0xFF19);               //650
    dfu_ctrl    (device, 0x03, 0x0018, 0xFF05);               //653
    
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF2A, buffer, 0x01); //656  > 0x00
    dfu_ctrl    (device, 0x03, 0x0004, 0xFF04);               //659
    dfu_ctrl    (device, 0x03, 0x0058, 0xFF67);               //662
    dfu_ctrl    (device, 0x03, 0x0018, 0xFF05);               //665
    dfu_ctrl    (device, 0x03, 0x0049, 0xFF1B);               //668
    dfu_ctrl    (device, 0x03, 0x0072, 0xFF1C);               //671
    dfu_ctrl    (device, 0x03, 0x00BB, 0xFF0C);               //674
    dfu_ctrl    (device, 0x03, 0x0031, 0xFF0D);               //677
    dfu_ctrl    (device, 0x03, 0x0024, 0xFF08);               //680
    
    fill_bloc(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x03); //683 09B1 FE    
    fill_first(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x2F); //686 0B16 0CD9 0D99 16 5F 20 40 2B33 2C20 35 83 38 88 3B 00 49 A0 4C FB 4D 93 5B 70 86 04 87  01 89 C8 93 DA 9C 57 A69D A701 A8D2 19 36 FE
    fill_bloc(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x03); //689 09B1 FE
    fill_second(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x37); //692 0E32 0FA2 108F 1160 16 5F 2B 43 44C0 4501 469A 49 B0 5C00 5D02 6A47 6B84 6C99 6D4E 6E55 6FE8 7082 88 7E 8A 38 95 03 4D 93 A69D A701 A8D2 09F3 FE
    buffer[0]=0x2B; buffer[1]=0x33; buffer[2]=0x88; buffer[3]=0xFE; buffer[4]=0x09; buffer[5]=0xF3; buffer[6]=0xFE; 
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x07); //695 2B3388 FE 09F3 FE
    
    dfu_ctrl    (device, 0x03, 0x0000, 0xFF00);               //698
    dfu_ctrl    (device, 0x03, 0x0020, 0xFF00);               //701
    
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF01, buffer, 0x01); //704 >0x02 (semble etre recherche en cours)
    
    dfu_ctrl    (device, 0x04, 0x0001, 0x0009);               //707
    dfu_ctrl    (device, 0x03, 0x0001, 0xFF34);               //710
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF09, buffer, 0x01); //713
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF0A, buffer, 0x01); //716
    
    dfu_ctrl    (device, 0x03, 0x0000, 0xFF34);               //719
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF24, buffer, 0x01); //722 >0x00
    
    //dfu_ctrl_get(device, 0x03, 0x0000, 0xFF24, buffer, 0x01); //725
    //dfu_ctrl_get(device, 0x03, 0x0000, 0xFF2A, buffer, 0x01); //728
    
    dfu_ctrl    (device, 0x10, 0x7F01, 0x0102);               //731 enable bulk reception
		//dfu_ctrl    (device, 0x10, 0x7F01, 0x0202);               //   disable ?
		
		//* URB_FUNCTION_SYNC_RESET_PIPE_AND_CLEAR_STALL 0x82 *
		
		//bulk...
//		usleep(10*1000*1000);
		
		dfu_ctrl    (device, 0x04, 0x0000, 0x0009);               //745 
		
    dfu_ctrl_get(device, 0x03, 0x0000, 0xFF2A, buffer, 0x01); //748 
    dfu_ctrl    (device, 0x03, 0x0004, 0xFF04);               //751
    dfu_ctrl    (device, 0x03, 0x0058, 0xFF67);               //754
    dfu_ctrl    (device, 0x03, 0x0018, 0xFF05);               //757
    dfu_ctrl    (device, 0x03, 0x0049, 0xFF1B);               //760
    dfu_ctrl    (device, 0x03, 0x0072, 0xFF1C);               //763
    dfu_ctrl    (device, 0x03, 0x00BB, 0xFF0C);               //766
    dfu_ctrl    (device, 0x03, 0x0031, 0xFF0D);               //769
    dfu_ctrl    (device, 0x03, 0x0024, 0xFF08);               //772    
   fill_bloc(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x03); //775
   fill_first(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x2F); //778
   fill_bloc(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x03); //781
   fill_second(buffer);
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x37); //784
   buffer[0]=0x2B; buffer[1]=0x33; buffer[2]=0x88; buffer[3]=0xFE; buffer[4]=0x09; buffer[5]=0xF3; buffer[6]=0xFE; 
    dfu_ctrl_set(device, 0x21, 0x0000, 0x00C6, buffer, 0x07); //787
    dfu_ctrl    (device, 0x03, 0x0000, 0xFF00);               //790
    dfu_ctrl    (device, 0x03, 0x0020, 0xFF00);               //793
                                    
		
		//usleep(1000000);
		//result = dfu_ctrl(device, 0x04,	0x0001,	0x0200);
		
		//result = dfu_ctrl_set(device, 0x30,	0x0001,	0x01a0, buffer, 0x100); 
		
		//result = dfu_ctrl(device, 0x20, 0xA055, 0xA044); //301
		
		result = dfu_ctrl_get(device, 0x20,	0x0000,	0x01a0, buffer, 0x200); //return 1689 ..		
		
 //   fprintf( stderr, "%s: F C020 0 01A0 B 1A >%x err=%d %s\n", __FUNCTION__, result, 0-result, usb_strerror());
	//	printdump(&buffer[0],0x200);
		


status->bStatus = DFU_STATUS_OK;
status->bState  = STATE_DFU_IDLE;
//exit(0);
return result;
}

/*
 *  DFU_UPLOAD Request (DFU Spec 1.0, Section 6.2)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *  length    - the maximum number of bytes to receive from the USB
 *              device - must be less than wTransferSize
 *  data      - the buffer to put the received data in
 *
 *  returns the number of bytes received or < 0 on error
 */
int dfu_upload( struct usb_dev_handle *device,
                const unsigned short address,
                const unsigned short length,
                char* data )
{
    int status;

    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    /* Sanity checks */
    if( (0 == length) || (NULL == data) ) {
        if( 0 != dfu_debug_level )
            fprintf( stderr,
                     "%s: data was NULL, or length is 0\n",
                     __FUNCTION__ );
        return -1;
    }

//		status = dfu_ctrl_get(device, 0x00,	0x0000,	address, data, length);

    status = usb_control_msg( device,
          /* bmRequestType */ 0xC0, // USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ 0,
          /* wValue        */ 0,
          /* wIndex        */ address,
          /* Data          */ data,
          /* wLength       */ length,
                              dfu_timeout );
    if( status < 0 ) {
        fprintf( stderr, "%s: error %d\n", __FUNCTION__, status );
    }

    return status;
}

/*
 *  DFU_GETSTATUS Request (DFU Spec 1.0, Section 6.1.2)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *  status    - the data structure to be populated with the results
 *
 *  return the number of bytes read in or < 0 on an error
 */
int dfu_get_status( struct usb_dev_handle *device,
                    const unsigned short interface,
                    struct dfu_status *status )
{
		char descriptor[0x50];

    char buffer[6];
    int result;

    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

/*
		result = usb_set_altinterface(device, 0);
    fprintf( stderr, "usb_set_altinterface result %x %s\n", result, usb_strerror());
    usleep(50000);
    */
    /* Initialize the status data structure */
    status->bStatus       = DFU_STATUS_ERROR_UNKNOWN;
    status->bwPollTimeout = 0;
    status->bState        = STATE_DFU_ERROR;
    status->iString       = 0;
/*
    usleep(50000);
    memset(buffer,0,8); 
    buffer[0] = 0x80; 
    buffer[2] = 
    result = usb_bulk_write( device, 0x0, buffer, 8, dfu_timeout );
                  usleep(50000);  
*/    
#if 0
    result = usb_control_msg( device,
          /* bmRequestType */ 0x21,//USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ 0x0A, //DFU_GETSTATUS,
          /* wValue        */ 0,
          /* wIndex        */ 0,
          /* Data          */ 0,
          /* wLength       */ 0,
                              dfu_timeout );

    
    fprintf( stderr, "%s: F 210A 0 0 0 0 >%x 0x%x %s\n", __FUNCTION__, 0-result, buffer[0] & 0xff, usb_strerror());
		usleep(50000);
#endif

		buffer[0] = 0x85;
    result = usb_control_msg( device,
          /* bmRequestType */ 0xC0, //USB_ENDPOINT_IN, //0x21,//USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ 1, //DFU_GETSTATUS,
          /* wValue        */ 0,
          /* wIndex        */ 1,
          /* Data          */ buffer,
          /* wLength       */ 2,
                              dfu_timeout );

		status->bState  = STATE_DFU_IDLE;		
		
		if (result==2) {
		  status->bStatus = DFU_STATUS_OK;
		  if (buffer[1]==0x01)
			  status->bState = DFU_STATE_dfuDNLOAD_IDLE;
			  fprintf( stderr, "%s: firmware present\n", __FUNCTION__);
      result=0;
		} else {
			//printdump(buffer,2);
	    fprintf( stderr, "%s: F C001 0 1 B 1 >%x buffer=0x%x %s\n", __FUNCTION__, result, buffer[0] & 0xff, usb_strerror());
		}
		usleep(50000);

    //inutile ?
    result = usb_release_interface(device, 1);
		result = usb_claim_interface(device, 1);  
		
    result = usb_get_descriptor(device, 2, 1, descriptor, 0x19);
    printdump(descriptor,0x19);
    fprintf( stderr, "usb_get_descriptor result %d ## Class=%x ## %s\n", result, descriptor[14], usb_strerror());
    usleep(50000);
    
		return result;
/*
    if( 6 == result ) {
        status->bStatus = buffer[0];
        status->bwPollTimeout = ((0xff & buffer[3]) << 16) |
                                ((0xff & buffer[2]) << 8)  |
                                (0xff & buffer[1]);

        status->bState  = buffer[4];
        status->iString = buffer[5];
    }*/
    return result;
}


int dfu_power( struct usb_dev_handle *device,
               const unsigned short interface )
{
    char buffer[2048];
    int result;

    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

#if 0

		dfu_ctrl_get(device, 0x00, 0, 0, buffer, 2048);
    dfu_ctrl_set(device, 0x00, 0, 0, buffer, 2048);

		dfu_ctrl_get(device, 0x00, 0, 2048, buffer, 2048);
		dfu_ctrl_set(device, 0x00, 0, 2048, buffer, 2048);

		dfu_ctrl_get(device, 0x00, 0, 4096, buffer, 2048); 
		buffer[16]=0xFA; //@0x1010 = 4112 = 4096 + 16
		dfu_ctrl_set(device, 0x00, 0, 4096, buffer, 2048);

		dfu_ctrl_get(device, 0x00, 0, 4096+2048, buffer, 965);
		dfu_ctrl_set(device, 0x00, 0, 4096+2048, buffer, 965);
		
		buffer[0]=0xFA;
		dfu_ctrl_set(device, 0x00, 0, 0x1E1A, buffer, 1);
    dfu_ctrl_set(device, 0x00, 0, 0x1010, buffer, 1);

		//500 mA = 0xFA * 2mA
    //buffer[0]=0xFA; buffer[1]=0xFA; buffer[2]=0xFA; buffer[3]=0xFA;
    dfu_ctrl_get(device, 0x00, 0, 0x1010, buffer, 0x04);
		dfu_ctrl_get(device, 0x00, 0, 0x1E1A, buffer, 0x04);
    
#endif

#if 0

		//0x81 - 0x03 0x01 0x01 - disable auto HID
		result = dfu_ctrl_get(device, 0x00, 0, 0x1e20, buffer, 3);
		if (buffer[0]==3) buffer[0]=0xFF;
		result = dfu_ctrl_set(device, 0x00, 0, 0x1e20, buffer, 3);

		//buffer[0]=0x12;buffer[1]=0x34;buffer[2]=0x56;
		//result = dfu_ctrl_get(device, 0x00, 0, 0x1260, buffer, 3);
		//result = dfu_ctrl_set(device, 0x00, 0, 0x1260, buffer, 3);
		
		//0x82 - 0xFF 0xFF 0xFF
		result = dfu_ctrl_get(device, 0x00, 0, 0x1e39, buffer, 3);
    result = dfu_ctrl_set(device, 0x00, 0, 0x1e39, buffer, 3);

		result = dfu_ctrl_read( device, 0x81, 0x06, 0x2200, 0, buffer, 63);

#endif

    //fprintf( stderr, "%s: reset device\n", __FUNCTION__);
		//result = usb_reset(device);
    //buffer[0]=0;
		
    return result;
}

/*
 *  DFU_CLRSTATUS Request (DFU Spec 1.0, Section 6.1.3)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *
 *  return 0 or < 0 on an error
 */
int dfu_clear_status( struct usb_dev_handle *device,
                      const unsigned short interface )
{
    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    return usb_control_msg( device,
        /* bmRequestType */ USB_ENDPOINT_OUT| USB_TYPE_CLASS | USB_RECIP_INTERFACE,
        /* bRequest      */ DFU_CLRSTATUS,
        /* wValue        */ 0,
        /* wIndex        */ interface,
        /* Data          */ NULL,
        /* wLength       */ 0,
                            dfu_timeout );
}


/*
 *  DFU_GETSTATE Request (DFU Spec 1.0, Section 6.1.5)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *  length    - the maximum number of bytes to receive from the USB
 *              device - must be less than wTransferSize
 *  data      - the buffer to put the received data in
 *
 *  returns the state or < 0 on error
 */
int dfu_get_state( struct usb_dev_handle *device,
                   const unsigned short interface )
{
    int result;
    char buffer[1];

    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    result = usb_control_msg( device,
          /* bmRequestType */ USB_ENDPOINT_IN | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
          /* bRequest      */ DFU_GETSTATE,
          /* wValue        */ 0,
          /* wIndex        */ interface,
          /* Data          */ buffer,
          /* wLength       */ 1,
                              dfu_timeout );

    /* Return the error if there is one. */
    if( result < 1 ) {
        return result;
    }

    /* Return the state. */
    return buffer[0];
}


/*
 *  DFU_ABORT Request (DFU Spec 1.0, Section 6.1.4)
 *
 *  device    - the usb_dev_handle to communicate with
 *  interface - the interface to communicate with
 *
 *  returns 0 or < 0 on an error
 */
int dfu_abort( struct usb_dev_handle *device,
               const unsigned short interface )
{
    if( 0 != dfu_verify_init(__FUNCTION__) )
        return -1;

    return usb_control_msg( device,
        /* bmRequestType */ USB_ENDPOINT_OUT | USB_TYPE_CLASS | USB_RECIP_INTERFACE,
        /* bRequest      */ DFU_ABORT,
        /* wValue        */ 0,
        /* wIndex        */ interface,
        /* Data          */ NULL,
        /* wLength       */ 0,
                            dfu_timeout );
}


char* dfu_state_to_string( int state )
{
    char *message = NULL;

    switch( state ) {
        case STATE_APP_IDLE:
            message = "appIDLE";
            break;
        case STATE_APP_DETACH:
            message = "appDETACH";
            break;
        case STATE_DFU_IDLE:
            message = "dfuIDLE";
            break;
        case STATE_DFU_DOWNLOAD_SYNC:
            message = "dfuDNLOAD-SYNC";
            break;
        case STATE_DFU_DOWNLOAD_BUSY:
            message = "dfuDNBUSY";
            break;
        case STATE_DFU_DOWNLOAD_IDLE:
            message = "dfuDNLOAD-IDLE";
            break;
        case STATE_DFU_MANIFEST_SYNC:
            message = "dfuMANIFEST-SYNC";
            break;
        case STATE_DFU_MANIFEST:
            message = "dfuMANIFEST";
            break;
        case STATE_DFU_MANIFEST_WAIT_RESET:
            message = "dfuMANIFEST-WAIT-RESET";
            break;
        case STATE_DFU_UPLOAD_IDLE:
            message = "dfuUPLOAD-IDLE";
            break;
        case STATE_DFU_ERROR:
            message = "dfuERROR";
            break;
    }

    return message;
}

/* Chapter 6.1.2 */
static const char *dfu_status_names[] = {
	[DFU_STATUS_OK]			= "No error condition is present",
	[DFU_STATUS_errTARGET]		= 
		"File is not targeted for use by this device",
	[DFU_STATUS_errFILE]		=
		"File is for this device but fails some vendor-specific test",
	[DFU_STATUS_errWRITE]		=
		"Device is unable to write memory",
	[DFU_STATUS_errERASE]		=
		"Memory erase function failed",
	[DFU_STATUS_errCHECK_ERASED]	=
		"Memory erase check failed",
	[DFU_STATUS_errPROG]		=
		"Program memory function failed",
	[DFU_STATUS_errVERIFY]		=
		"Programmed emmory failed verification",
	[DFU_STATUS_errADDRESS]		=
		"Cannot program memory due to received address that is out of range",
	[DFU_STATUS_errNOTDONE]		=
		"Received DFU_DNLOAD with wLength = 0, but device does not think that it has all data yet",
	[DFU_STATUS_errFIRMWARE]	=
		"Device's firmware is corrupt. It cannot return to run-time (non-DFU) operations",
	[DFU_STATUS_errVENDOR]		=
		"iString indicates a vendor specific error",
	[DFU_STATUS_errUSBR]		=
		"Device detected unexpected USB reset signalling",
	[DFU_STATUS_errPOR]		=
		"Device detected unexpected power on reset",
	[DFU_STATUS_errUNKNOWN]		=
		"Something went wrong, but the device does not know what it was",
	[DFU_STATUS_errSTALLEDPKT]	=
		"Device stalled an unexpected request",
};


const char *dfu_status_to_string(int status)
{
	if (status > DFU_STATUS_errSTALLEDPKT)
		return "INVALID";
	return dfu_status_names[status];
}

